import java.util.TreeMap;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Arrays;

import processing.core.*; 

class pluginLines {
  PApplet myPApplet;
  Song mySong;
  PFont myFont;
  int targetWidth;

  public PVector eye, target, nextEye, nextTarget, syllableTarget;
  public float aimFrame, nextAimFrame;

  float seedX, seedY, seedZ;

  pluginLines (PApplet newPApplet, Song newSong, PFont newFont, int newTargetWidth) {
    try {
      mySong = (Song)ObjectCloner.deepCopy(newSong);
    }
    catch(Exception e)
    {
      System.out.println("Exception in main = " +  e);
    }
    myPApplet = newPApplet;

    myFont = newFont;
    targetWidth = newTargetWidth;

    eye = new PVector();
    target = new PVector();
    aimFrame = 0;
    nextEye = new PVector();
    nextTarget = new PVector(); 
    nextAimFrame = 0;
    syllableTarget = new PVector();

    seedX = myPApplet.random(12);
    seedY = myPApplet.random(12);
    seedZ = myPApplet.random(12);

    myPApplet.resetMatrix();
  }



  void draw(int currentFrame, int textColor, PMatrix3D songMatrix, PMatrix3D lineMatrix,  boolean stepWidth, boolean stepHeight, boolean randomRotate, boolean dropShadow, int[] syllableAnimation) {  
    myPApplet.pushMatrix();
    myPApplet.applyMatrix(songMatrix);    

    TreeMap songLines = mySong.getLines();  
    Iterator<Integer> lineItr = songLines.keySet().iterator();

    float lineOffsetX = 0;
    myPApplet.pushMatrix();
    while (lineItr.hasNext()) {
      int curLineKey = lineItr.next();
      Line curLine = (Line)songLines.get(curLineKey);

      if (stepWidth) myPApplet.translate(lineOffsetX,0);      
      if (stepHeight) myPApplet.translate(0,myPApplet.textAscent());

      if (randomRotate) {
        myPApplet.rotateX(seedX * ((curLine.getStartFrame()/2) - curLine.getStartFrame() ));
        myPApplet.rotateY(seedY *  ((curLine.getStartFrame()/2) - curLine.getStartFrame() ));
        myPApplet.rotateZ(seedZ *  ((curLine.getStartFrame()/2) - curLine.getStartFrame() ));
      }


      if (curLine.getStartFrame() > currentFrame){    
        myPApplet.pushMatrix();
        nextTarget.set(myPApplet.modelX(0,0,0),myPApplet.modelY(0,0,0),myPApplet.modelZ(0,0,0));
        nextEye.set(myPApplet.modelX(0,0,500),myPApplet.modelY(0,0,500),myPApplet.modelZ(0,0,500));
        nextAimFrame = curLine.getStartFrame();
        myPApplet.popMatrix();        
        break;                
      }



      myPApplet.pushMatrix();
      target.set(myPApplet.modelX(0,0,0),myPApplet.modelY(0,0,0),myPApplet.modelZ(0,0,0));
      eye.set(myPApplet.modelX(0,0,500),myPApplet.modelY(0,0,500),myPApplet.modelZ(0,0,500));
      aimFrame = curLine.getStartFrame();      
      myPApplet.popMatrix();    

      myPApplet.applyMatrix(lineMatrix);

      TreeMap songWords = curLine.getWords();
      Iterator<Integer> wordItr = songWords.keySet().iterator();

      float wordOffsetX = 0;
      myPApplet.pushMatrix();
      while (wordItr.hasNext()) {
        int curWordKey = wordItr.next();
        Word curWord = (Word)songWords.get(curWordKey);
        myPApplet.translate(wordOffsetX,0);

        if (curWord.getStartFrame() > currentFrame)
          continue;      


        TreeMap songSyllables = curWord.getSyllables();
        Iterator<Integer> syllableItr = songSyllables.keySet().iterator();

        float syllableOffsetX = 0;
        myPApplet.pushMatrix();
        while (syllableItr.hasNext()) {
          int curSyllableKey = syllableItr.next();
          Syllable curSyllable = (Syllable)songSyllables.get(curSyllableKey);       

          if (curSyllable.getStartFrame() > currentFrame)
            continue;

          if (currentFrame - curSyllable.getStartFrame() < 8 ) {
            float aniFactor = (currentFrame - curSyllable.getStartFrame()) / 8f;
            if (syllableAnimation[0] == 1)
              myPApplet.translate(0,0,myPApplet.round(256 - ( aniFactor* 256)));
            if (syllableAnimation[1] == 1)
              myPApplet.rotateX(myPApplet.radians(myPApplet.round(85 - ( aniFactor* 85))));
            if (syllableAnimation[2] == 1)
              myPApplet.scale(aniFactor);
          }  


          myPApplet.translate(syllableOffsetX,0);


          myPApplet.textFont(myFont);

          if (dropShadow){
            myPApplet.shininess(0);  
            myPApplet.fill(0,(float)(curLine.getStartFrame() * 1.0 / currentFrame) * 128) ;
            myPApplet.pushMatrix();
            myPApplet.translate(targetWidth/80,targetWidth/80,-targetWidth/80);
            myPApplet.text(curSyllable.getLetters(), 0, 0);  
            myPApplet.popMatrix();
          }          

          myPApplet.shininess(100);  
          myPApplet.fill(textColor , (float)(curLine.getStartFrame() * 1.0 / currentFrame) * 255) ;
          myPApplet.text(curSyllable.getLetters(), 0, 0);

          syllableTarget.set(myPApplet.modelX(0,0,0), myPApplet.modelY(0,0,0), myPApplet.modelZ(0,0,0));

          syllableOffsetX = myPApplet.textWidth(curSyllable.getLetters());


        }
        myPApplet.popMatrix();
        wordOffsetX = myPApplet.textWidth(curWord.getLetters());        
      }

      lineOffsetX = myPApplet.textWidth(curLine.getLetters() + " ");
      if (!lineItr.hasNext()) {
        myPApplet.pushMatrix();
        if (stepWidth) myPApplet.translate(lineOffsetX/2,0);      
        if (stepHeight) myPApplet.translate(0,curLine.getHeight());        
        nextTarget.set(myPApplet.modelX(0,0,0),myPApplet.modelY(0,0,0),myPApplet.modelZ(0,0,0));
        nextEye.set(myPApplet.modelX(0,0,500),myPApplet.modelY(0,0,500),myPApplet.modelZ(0,0,500));
        nextAimFrame = curLine.getEndFrame();
        myPApplet.popMatrix();            
      }        
      myPApplet.popMatrix();

    }   

    myPApplet.popMatrix();
    myPApplet.popMatrix();

  }

  PVector getAim(int currentFrame) {
    int curAimX, curAimY, curAimZ;
    if(nextAimFrame - aimFrame > 300) {
      if (currentFrame < ((nextAimFrame - aimFrame)/2) + aimFrame) {
        nextTarget.set(0,0,0);
      } 
      else {
        target.set(0,0,0);
      }
    }
    if (currentFrame > nextAimFrame) {
      curAimX = myPApplet.floor(nextTarget.x);
      curAimY = myPApplet.floor(nextTarget.y);
      curAimZ = myPApplet.floor(nextTarget.z);
    }
    else {
      curAimX = myPApplet.floor(target.x + (nextTarget.x - target.x) * ((currentFrame - aimFrame) / (nextAimFrame - aimFrame)));          
      curAimY = myPApplet.floor(target.y + (nextTarget.y - target.y) * ((currentFrame - aimFrame) / (nextAimFrame - aimFrame)));
      curAimZ = myPApplet.floor(target.z + (nextTarget.z - target.z) * ((currentFrame - aimFrame) / (nextAimFrame - aimFrame)));        
    }

    return(new PVector(curAimX, curAimY, curAimZ));
  }
}


















